import copy
import os
import pickle
import subprocess
import sys

import pyinotify

import nose

from runnynose.tracker import DependencyTracker
from runnynose.plugin import COMM_FILE
from runnynose.plugin import TestCollectorPlugin
from runnynose.plugin import TestContextPlugin
from runnynose.util import NotSpecified

class TestRunner(object):
    
    def __init__(self, argv):
        self._argv = argv
        self._tracker = DependencyTracker.get()
        self._orphans = set()
    
    def start(self):
        self._run()
    
    def handle_changed(self, pathname):
        test_addresses = self._collect(pathname)
        for address in test_addresses:
            if address[2]:
                self._tracker.add(address[0], address)
        
        old_tests = set([test for test in self._tracker.dependencies_of(pathname)
                         if test[0] == pathname])
        new_tests = test_addresses
        deleted_tests = old_tests - new_tests
        for test in deleted_tests:
            self._tracker.remove_test(test)
        
        tests = self._tracker.dependencies_of(pathname)
        if tests:
            tests.update(self._orphans)
            test_ids = [':'.join(address[1:]) for address in tests]
            self._run(list(test_ids))
    
    def handle_deleted(self, pathname):
        tests = self._tracker.dependencies_of(pathname)
        deleted_tests = set()
        for address in tests:
            if address[0] == pathname:
                deleted_tests.add(address)
                if address in self._orphans:
                    self._orphans.remove(address)
        for address in deleted_tests:
            self._tracker.remove_test(address)
        
        tests = self._tracker.dependencies_of(pathname)
        self._orphans.update(tests)
        if tests:
            test_ids = [':'.join(address[1:]) for address in tests]
            self._run(list(test_ids))
    
    def _run(self, test_ids=NotSpecified):
        if test_ids is NotSpecified:
            test_ids = []
        
        argv = self._argv + ['--with-runnynose-test-context'] + test_ids
        argv[0] = 'nosetests'
        subprocess.call(argv)
        with open(COMM_FILE, 'r') as f:
            result = pickle.load(f)
        self._tracker.load(result['dependencies'])
        for test in result['passed']:
            if test in self._orphans:
                self._orphans.remove(test)
    
    def _collect(self, pathname):
        with open('/dev/null', 'w') as nowhere:
            argv = self._argv+['--collect-only', '--with-runnynose-test-collector'] + [pathname]
            argv[0] = 'nosetests'
            subprocess.call(argv, stdout=nowhere, stderr=nowhere)
        with open(COMM_FILE, 'r') as f:
            test_ids = pickle.load(f)
        return test_ids

class Watcher(pyinotify.ProcessEvent):
    
    def my_init(self, runner, argv):
        self._runner = runner
        self._argv = copy.copy(argv)
    
    def _is_python_source(self, pathname):
        basename = os.path.basename(pathname)
        if basename.startswith('.#'):
            return False
        if basename.endswith('.py'):
            return True
        else:
            return False
    
    def process_IN_CLOSE_WRITE(self, event):
        if self._is_python_source(event.pathname):
            print 'CHANGED: ', event.pathname
            self._runner.handle_changed(event.pathname)
    
    def process_IN_DELETE(self, event):
        if self._is_python_source(event.pathname):
            print 'DELETED:', event.pathname
            self._runner.handle_deleted(event.pathname)

def main(argv=NotSpecified):
    if argv is NotSpecified:
        argv = sys.argv
    
    runner = TestRunner(argv)
    runner.start()
    
    # print DependencyTracker.get().__dict__
    wm = pyinotify.WatchManager()
    watcher = Watcher(runner=runner, argv=argv)
    notifier = pyinotify.Notifier(wm, default_proc_fun=watcher)
    wm.add_watch(os.getcwd(), pyinotify.ALL_EVENTS, rec=True, auto_add=True)
    notifier.loop()




from astkit import ast

from runnynose.tracker import get_record
from runnynose.tracker import get_setup

class TrackingTransformer(ast.NodeTransformer):
    
    def process(self, module):
        return self.visit(module)
    
    def visit_Module(self, module):
        self.generic_visit(module)
        setup = get_setup()
        if module.body:
            for node in setup:
                ast.copy_location(node, module.body[0])
            
            futures = []
            while (isinstance(module.body[0], ast.ImportFrom) 
                   and (module.body[0].module == '__future__')):
                futures.append(module.body.pop(0))
            module.body = futures + setup + module.body
        else:
            module.body = setup
        return module
    
    def visit_FunctionDef(self, defn):
        self.generic_visit(defn)
        record = get_record()
        for node in record:
            ast.copy_location(node, defn.body[0])
        defn.body = record + defn.body
        return defn



import inspect

from astkit import ast

def get_setup():
    source = inspect.getsource(DependencyTracker.get_tracker)
    source = "\n".join([line[4:] for line in source.splitlines()])
    mod = ast.parse(source)
    defn = mod.body[0]
    setup = defn.body[:]
    for stmt in setup:
        stmt.lineno -= 1
    return setup

def get_record():
    source = inspect.getsource(DependencyTracker.record_use)
    source = "\n".join([line[4:] for line in source.splitlines()])
    mod = ast.parse(source)
    defn = mod.body[0]
    record = defn.body[:]
    for stmt in record:
        stmt.lineno -= 1
    return record
    
class DependencyTracker(object):
    current_test = None
    
    _instance = None
    @classmethod
    def get(cls):
        if cls._instance is None:
            cls._instance = cls()
        return cls._instance
    
    def __init__(self):
        self.reset()
    
    def reset(self):
        self._dependencies = {}
    
    def add(self, pathname, address):
        if address:
            addresses = self._dependencies.setdefault(pathname, set())
            addresses.add(address)
    
    def dependencies_of(self, pathname):
        return self._dependencies.get(pathname, set())
    
    def depends_on(self, address):
        pathnames = set()
        for pathname, addresses in self._dependencies.items():
            if address in addresses:
                pathnames.add(pathname)
        return pathnames
    
    def remove_test(self, address):
        for pathname in self.depends_on(address):
            self._dependencies[pathname].remove(address)
    
    # Dumping and loading methods
    def dump(self):
        return self._dependencies
    
    def load(self, dependencies):
        for pathname, addresses in dependencies.items():
            for address in addresses:
                self.add(pathname, address)
    
    # These are for instrumented code to call
    def record_pathname(self, pathname):
        self.add(pathname, self.current_test)
    
    @classmethod
    def get_tracker(cls):
        from runnynose.tracker import DependencyTracker
        tracker = DependencyTracker.get()
    
    @classmethod
    def record_use(cls):
        tracker.record_pathname(__file__)



from pickle import dump

from astkit import install_processor

from nose.plugins import Plugin

from runnynose.hook import TrackingTransformer
from runnynose.tracker import DependencyTracker

COMM_FILE = '.runnynose'

class TestContextPlugin(Plugin):
    name='runnynose-test-context'
    
    def begin(self):
        install_processor(TrackingTransformer())
        self._passed = set()
    
    def startTest(self, test):
        pathname = test.address()[0]
        current_address = test.address()
        DependencyTracker.get().add(pathname, current_address)
        DependencyTracker.current_test = current_address
        self._passed.add(current_address)
    
    def handleError(self, test, err):
        self._passed.remove(test.address())
    handleFailure = handleError
    
    def stopTest(self, test):
        DependencyTracker.current_test = None
        
    def report(self, stream):
        result = {'dependencies': DependencyTracker.get().dump(),
                  'passed': self._passed}
        with open(COMM_FILE, 'w') as f:
            dump(result, f)
    
class TestCollectorPlugin(Plugin):
    name='runnynose-test-collector'
    
    def __init__(self, *args, **kwargs):
        super(TestCollectorPlugin, self).__init__(*args, **kwargs)
        self.tests = set()
    
    def startTest(self, test):
        self.tests.add(test.address())
    
    def report(self, stream):
        with open(COMM_FILE, 'w') as f:
            dump(self.tests, f)
        return True



from setuptools import setup, find_packages
import sys, os

version = '0.1.0'

setup(name='runnynose',
      version=version,
      description="A continuous nose runner that tracks dependencies",
      long_description="""\
""",
      classifiers=[], # Get strings from http://pypi.python.org/pypi?%3Aaction=list_classifiers
      keywords='',
      author='Matthew J Desmarais',
      author_email='matthew.desmarais@gmail.com',
      url='',
      license='MIT',
      packages=find_packages(exclude=['ez_setup', 'examples', 'tests']),
      include_package_data=True,
      zip_safe=False,
      install_requires=[
          'astkit>=0.5.2'
      ],
      entry_points="""
      [console_scripts]
      runnynose = runnynose.cmd:main
      
      [nose.plugins.0.10]
      collector = runnynose.plugin:TestCollectorPlugin
      context = runnynose.plugin:TestContextPlugin
      """,
      )
